#include "bleaf.hh"
#include "bnode.hh"
#include "cloth.hh"
#include "common.hh"
#include <cmath>
#include <cstdlib>

using namespace std;

Cloth::Cloth (istream& in, string version, int ixSize, int iySize, double idist,
              double id2, double istrength, double is2, double icurve, double iweight,
              double hNoise, double vNoise, double ielongation, double idamping,
              bool iinters): xSize(ixSize), ySize(iySize), points(new Vertex[ixSize*iySize]),
                             dist(idist), d2(id2), strength(istrength), s2(is2), curve(icurve),
                             weight(iweight), elongation(ielongation), damping(idamping),
                             inters (iinters) {
  int major, minor, i, j;
  i = version.find_first_of ("0123456789");
  j = version.find ('.', i);
  major = atoi (version.substr (i, j-i).c_str());
  i = version.find_first_of ("0123456789", j);
  j = version.find ('.', i);
  minor = atoi (version.substr (i, j-i).c_str());

  if ((major == 0) && (minor == 2)) {
    for (int i = 0; i < xSize; i++)
      for (int j = 0; j < ySize; j++) {
        double x, y, z;
        in >> x;
        in >> y;
        in >> z;
        point (i, j) = Vertex (Vector (x, y, z), this);
      }
  } else { // The file wasn't generated by version 0.1.x of the simulator,
    // assume it uses the current format
    for (int i = 0; i < xSize; i++)
      for (int j = 0; j < ySize; j++) {
        double x, y, z;
        in >> x;
        in >> y;
        in >> z;
        point (i, j) = Vertex (Vector (x, y, z), this);
        in >> x;
        in >> y;
        in >> z;
        point (i, j).speed = Vector (x, y, z);
      }
  }
  initLinks();
}

Cloth::Cloth (int ixSize, int iySize, double idist, double id2,
              double istrength, double is2, double icurve, double iweight,
              double hNoise, double vNoise, double ielongation, double idamping,
              bool iinters): xSize(ixSize), ySize(iySize), points(new Vertex[ixSize*iySize]),
                             dist(idist), d2(id2), strength(istrength), s2(is2), curve(icurve),
                             weight(iweight), elongation(ielongation), damping(idamping),
                             inters(iinters) {
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      point (i, j) = Vertex (Vector ((i-xSize/2)*dist*(1-hNoise-vNoise)+(random()/(double)RAND_MAX)*dist*hNoise, 
                                     (j-ySize/2)*dist*(1-hNoise-vNoise)+(random()/(double)RAND_MAX)*dist*hNoise,
                                     (random()/(double)RAND_MAX)*vNoise), this);
  // I still need to link them to their neighbors...
  initLinks();
}

Cloth::Cloth (int ixSize, int iySize, double idist, double id2,
              double istrength, double is2, double icurve, double iweight,
              double hNoise, double vNoise, double ielongation, double idamping,
              bool iinters, const Vector& orig, const Vector& dir1, const Vector& dir2):
  xSize(ixSize), ySize(iySize), points(new Vertex[ixSize*iySize]),
  dist(idist), d2(id2), strength(istrength), s2(is2), curve(icurve),
  weight(iweight), elongation(ielongation), damping(idamping), inters(iinters) {

  Vector n = dir1.cross (dir2).normalize();
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      point (i, j) = Vertex (orig +
                             ((i+(random()/(double)RAND_MAX)*hNoise)/((double)(xSize-1)))*dir1 +
                             ((j+(random()/(double)RAND_MAX)*hNoise)/((double)(ySize-1)))*dir2 +
                             (random()/(double)RAND_MAX)*vNoise*n,
                             this);
  // I still need to link them to their neighbors...
  initLinks();
}

void Cloth::initLinks() {
  // neighbors
  for (int i = 0; i < xSize-1; i++)
    for (int j = 0; j < ySize-1; j++) {
      point (i, j).addNeighbor (&point (i+1, j));
      point (i, j).addNeighbor (&point (i, j+1));
    }
  for (int i = 0; i < xSize-1; i++)
    point (i, ySize-1).addNeighbor (&point (i+1, ySize-1));
  // neighbors and corners
  for (int j = 0; j < ySize-1; j++) {
    point (xSize-1, j).addNeighbor (&point (xSize-1, j+1));
    point (xSize-1, j).addCorner (&point (xSize-2, j+1));
  // corners
    point (0, j).addCorner (&point (1, j+1));
  }
  for (int i = 1; i < xSize-1; i++)
    for (int j = 0; j < ySize-1; j++) {
      point (i, j).addCorner (&point (i-1, j+1));
      point (i, j).addCorner (&point (i+1, j+1));
    }
  // curvings
  for (int i = 1; i < xSize-1; i++) {
    point (i, 0).addCurving (&point (i-1, 0), &point (i+1, 0));
    point (i, ySize-1).addCurving (&point (i-1, ySize-1), &point (i+1, ySize-1));
  }
  for (int j = 1; j < ySize-1; j++) {
    point (0, j).addCurving (&point (0, j-1), &point (0, j+1));
    point (xSize-1, j).addCurving (&point (xSize-1, j-1), &point (xSize-1, j+1));
  }
  for (int i = 1; i < xSize-1; i++)
    for (int j = 1; j < ySize-1; j++) {
      point (i, j).addCurving (&point (i-1, j), &point (i+1, j));
      point (i, j).addCurving (&point (i, j-1), &point (i, j+1));
    }
}
                                 
Cloth::~Cloth() {
  delete[] points; points = (Vertex*)-1;
  if (inters)
    delete boundingTree; boundingTree = (BBox*)-1;
}

void Cloth::translate (const Vector& v) {
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      point (i, j).newloc = point (i, j).location_ = point (i, j).location_ + v;
}

void Cloth::rotate (const Vector& v) {
  Vector coefs[3];
  double cx = cos (v.x()), cy = cos (v.y()), cz = cos (v.z());
  double sx = sin (v.x()), sy = sin (v.y()), sz = sin (v.z());


  coefs[0] = Vector (cy*cz, sx*sy*cz-cx*sz, cx*sy*cz+sx*sz);
  coefs[1] = Vector (cy*sz, sx*sy*sz+cx*cz, cx*sy*sz-sx*cz);
  coefs[2] = Vector (  -sy,          sx*cy,          cx*cy);

  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      point (i, j).newloc = point (i, j).location_ =
        Vector (point (i, j).location_.dot (coefs[0]),
                point (i, j).location_.dot (coefs[1]),
                point (i, j).location_.dot (coefs[2]));
}

void Cloth::simulate (int steps) {
  if (inters) {
    cerr << endl << "Creating self-intersection bounding tree...";
    createBoundingTree();
  }
  maxDist = dist*(1+elongation);
  maxD2 = d2*(1+elongation);
  cerr << endl << "Simulating..." << endl;
  for (int i = 0; i < steps; i++) {
    computeNormals();
    computeForces();
    move();
    if ((i & 0xF) == 0)
      if ((i & 0x3FF) != 0)
        cerr << ".";
      else
        cerr << " " << i << endl;
  }
  cerr << endl;
}

void Cloth::computeForces() {
  for (int i = xSize-1; i >= 0; i--)
    for (int j = ySize-1; j >=0; j--)
      point (i, j).computeForce();
}

void Cloth::move() {
  for (int i = xSize-1; i >= 0; i--)
    for (int j = ySize-1; j >= 0; j--) {
      point (i, j).move();
    }
  for (int i = xSize-1; i >= 0; i--)
    for (int j = ySize-1; j >=0; j--)
      point (i, j).adjustLength();
  for (int i = xSize-1; i >= 0; i--)
    for (int j = ySize-1; j >=0; j--)
      point (i, j).endAdjust();
#ifdef COMPUTE_INTERSECTIONS
  if (inters)
    intersect();
#endif
  for (int i = xSize-1; i >= 0; i--)
    for (int j = ySize-1; j >= 0; j--) {
      point (i, j).doMove();
      point (i, j).reset();
    }
}

#define CLOTH(X_, Y_) point (X_, Y_).location()

void Cloth::output_cyl (ostream& out) const {
  computeNormals();

  out << "union {" << endl;
  for (int i = 1; i < xSize; i++)
    for (int j = 1; j < ySize; j++) {
      double tension = ((CLOTH(i-1, j) - CLOTH(i, j)).length()-dist)/dist;
      if (tension < 0) tension = 0;
      else tension = tension/(2*elongation);
      if (tension > 1) tension = 1;
      double red, green;
      if (tension < 0.5) {
        red = 2*tension;
        green = 1;
      } else {
        red = 1;
        green = 2*(1-tension);
      }
      out << " cylinder { " << CLOTH(i-1, j) << ", " << CLOTH(i, j) << ", " << dist/10.0 << " pigment { red " << red << " green " << green << " } }" << endl;
      tension = ((CLOTH(i, j-1) - CLOTH(i, j)).length()-dist)/dist;
      if (tension < 0) tension = 0;
      else tension = tension/(2*elongation);
      if (tension > 1) tension = 1;
      if (tension < 0.5) {
        red = 2*tension;
        green = 1;
      } else {
        red = 1;
        green = 2*(1-tension);
      }
      out << " cylinder { " << CLOTH(i, j-1) << ", " << CLOTH(i, j) << ", " << dist/10.0 << " pigment { red " << red << " green " << green << " } }" << endl;
    }
  out << endl << "}" << endl;
}

void Cloth::output (ostream& out) const {
  computeNormals();

  out << "mesh {" << endl;
  for (int i = 1; i < xSize; i++) {
    for (int j = 1; j < ySize; j++) {
      out << "  smooth_triangle { " 
          << CLOTH(i, j) << ", " << normal (i, j) << ", "
          << CLOTH(i-1, j) << ", " << normal (i-1, j) << ", "
          << CLOTH(i-1, j-1) << ", " << normal (i-1, j-1)
          << " }" << endl;
      out << "  smooth_triangle { "
          << CLOTH(i, j) << ", " << normal (i, j) << ", "
          << CLOTH(i, j-1) << ", " << normal (i, j-1) << ", "
          << CLOTH(i-1, j-1) << ", " << normal (i-1, j-1)
          << " }" << endl;
      out << endl;
    }
    out << "/******************************/" << endl;
  }
  out << endl << "}" << endl;
}

void Cloth::output2 (ostream& out) const {
  computeNormals();

  out << "#version unofficial MegaPov 0.6;" << endl;
  out << "mesh2 {" << endl;
  out << "  vertex_vectors {" << endl << "    " << xSize*ySize;
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      out << "," << endl << "    " << CLOTH(i, j);
  out << endl << "  }" << endl;
  out << "  normal_vectors {" << endl << "    " << xSize*ySize;
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      out << "," << endl << "    " << normal (i, j);
  out << endl << "  }" << endl;
  out << "  uv_vectors {" << endl << "    " << xSize*ySize;
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      out << "," << endl << "    <" << i/(double)(xSize-1) << ", " << j/(double)(ySize-1) << ">";
  out << endl << "  }" << endl;
  out << "  face_indices {" << endl << "    " << 2*(xSize-1)*(ySize-1);
  for (int i = 1; i < xSize; i++)
    for (int j = 1; j < ySize; j++) {
      out << "," << endl;
      out << "    <" << j+i*ySize << ", " << j-1+i*ySize << ", " << j-1+(i-1)*ySize << ">, " << endl;
      out << "    <" << j+i*ySize << ", " << j+(i-1)*ySize << ", " << j-1+(i-1)*ySize << ">";
    }
  out << endl << "  }" << endl;
  out << endl << "}" << endl;
}

void Cloth::write (ostream& out) const {
  out << VERSION << endl;
  out << xSize << " " << ySize << endl;
  out.precision (20);
  for (int i = 0; i < xSize; i++) {
    for (int j = 0; j < ySize; j++)
      out << point (i, j).location().x() << " "
          << point (i, j).location().y() << " "
          << point (i, j).location().z() << " "
          << point (i, j).speed.x() << " "
          << point (i, j).speed.y() << " "
          << point (i, j).speed.z() << endl;
    out << endl;
  }
}

static inline void computeNormal (Vector& N, double L, const Vector& V11,
                                  const Vector& V12, const Vector& V21, const Vector& V22) {
  N = L*((V12)-(V11)).cross((V22)-(V21));
  if (N.l2() < EPSILON)
    N = (L/1.41421356237)*((V12)-(V21)).cross((V22)-(V11));
}
void Cloth::computeNormals() const {
                                // corners
  normal (0, 0) = (1.0/(dist*dist))*(CLOTH(1, 0)-CLOTH(0, 0))
    .cross (CLOTH(0, 1)-CLOTH(0, 0));
  normal (0, ySize-1) = (1.0/(dist*dist))*(CLOTH(1, ySize-1)-CLOTH(0, ySize-1))
    .cross (CLOTH(0, ySize-1)-CLOTH(0, ySize-2));
  normal (xSize-1, 0) = (1.0/(dist*dist))*(CLOTH(xSize-1, 0)-CLOTH(xSize-2, 0))
    .cross (CLOTH(xSize-1, 1)-CLOTH(xSize-1, 0));
  normal (xSize-1, ySize-1) = (1.0/(dist*dist))*(CLOTH(xSize-1, ySize-1)-CLOTH(xSize-2, ySize-1))
    .cross (CLOTH(xSize-1, ySize-1)-CLOTH(xSize-1, ySize-2));
                                // sides
  for (int j = 1; j < ySize-1; j++) {
    computeNormal (normal (0, j), (1.0/(2*dist*dist)), CLOTH(0, j), CLOTH(1, j),
                   CLOTH(0, j-1), CLOTH(0, j+1));
    computeNormal (normal (xSize-1, j), (1.0/(2*dist*dist)), CLOTH(xSize-2, j),
                   CLOTH(xSize-1, j), CLOTH(xSize-1, j-1), CLOTH(xSize-1, j+1));
  }
  for (int i = 1; i < xSize-1; i++) {
    computeNormal (normal (i, 0), (1.0/(2*dist*dist)), CLOTH(i-1, 0), CLOTH(i+1, 0),
                   CLOTH(i, 0), CLOTH(i, 1));
    computeNormal (normal (i, ySize-1), (1.0/(2*dist*dist)), CLOTH(i-1, ySize-1),
                   CLOTH(i+1, ySize-1), CLOTH(i, ySize-2), CLOTH(i, ySize-1));
  }
                                // middle
  for (int i = 1; i < xSize-1; i++)
    for (int j = 1; j < ySize-1; j++)
      computeNormal (normal(i, j), (1.0/(4*dist*dist)), CLOTH(i-1, j), CLOTH(i+1, j),
                     CLOTH(i, j-1), CLOTH(i, j+1));
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      if (normal (i, j).l2() < EPSILON) {
        cerr << endl << i << ", " << j << ": " << normal(i, j);
      }
}

void Cloth::attach (double u, double v, const Vector& where, const Vector& speed) {
  register int i = static_cast<int> (rint (u*(xSize-1)));
  register int j = static_cast<int> (rint (v*(ySize-1)));
  point (i, j).attach (where, speed);
}

void Cloth::intersect() {
  boundingTree->computeBounds();
  for (int i = 0; i < xSize; i++)
    for (int j = 0; j < ySize; j++)
      boundingTree->intersect (point(i, j));
}

void Cloth::createBoundingTree() {
  BBox** tmp = new BBox*[(xSize-1)*(ySize-1)];
  int w = xSize-1, l = ySize-1;
  for (int i = w-1; i >= 0; i--)
    for (int j = l-1; j >= 0; j--)
      tmp[j*w+i] = new BLeaf (point(i, j+1), point(i+1, j+1), point(i, j), point (i+1, j));

  while ((w > 1) || (l > 1)) {
    BNode* n;
    if ((w & 1) != 0) {
      int offset = (l & 1);
      for (int j = 0; j < l/2; j++) {
        n = new BNode();
        n->addChild (tmp[(j*2+offset)*(xSize-1)]);
        n->addChild (tmp[(j*2+offset+1)*(xSize-1)]);
        tmp[(j+offset)*(xSize-1)] = n;
      }
    }
    if ((l & 1) != 0) {
      int offset = (w & 1);
      for (int i = 0; i < w/2; i++) {
        n = new BNode();
        n->addChild (tmp[(i*2+offset)]);
        n->addChild (tmp[(i*2+offset+1)]);
        tmp[(i+offset)] = n;
      }
    }
    int xOffset = (w & 1), yOffset = (l & 1);
    for (int i = 0; i < w/2; i++)
      for (int j = 0; j < l/2; j++) {
        n = new BNode();
        n->addChild (tmp[(j*2+yOffset)*(xSize-1)+(i*2+xOffset)]);
        n->addChild (tmp[(j*2+yOffset+1)*(xSize-1)+(i*2+xOffset)]);
        n->addChild (tmp[(j*2+yOffset)*(xSize-1)+(i*2+xOffset+1)]);
        n->addChild (tmp[(j*2+yOffset+1)*(xSize-1)+(i*2+xOffset+1)]);
        tmp[(j+yOffset)*(xSize-1)+(i+xOffset)] = n;
      }

    w = (w+1)/2;
    l = (l+1)/2;
  }

  boundingTree = tmp[0];

  delete[] tmp;
}
