#include "common.hh"
#include "hf.hh"
#include <fstream>

using namespace std;

HeightField::HeightField (char* file): data(NULL), origin(O), size(1, 1, 1),
                                       length(0), width(0) {
  // Read a TGA file
  ifstream in (file);
  char header[18];

  in.read (header, 18);
  if (!in) {
    cerr << endl << "'" << file << "' is not a valid TGA file!" << endl;
    exit (1);
  }
  if ((header[2] == 0) || (header[2] > 3)) {
    cerr << endl << "Can only read uncompressed TGA files!" << endl;
    exit (1);
  }
  if (header[16] != 24) {
    cerr << endl << "Can't read TGAs with aplpha channel!" << endl;
    exit (1);
  }
  length = header[14] + (header[15]<<8);
  width  = header[12] + (header[13]<<8);
  data = new double[width*length];

  int psize = header[16]/8;     // pixel size
  long maxp = ~(-1 << (psize*8));

  in.ignore (header[5] + (header[6]<<8)); // skip color map

  for (int i = 0; i < width*length; i++) { // ignore file orientation!
    long pixel = 0;
    for (int k = 0; k < psize; k++) {
      unsigned char c = in.get();
      pixel = (pixel << 8) + c;
    }
    data[i] = pixel / static_cast<double> (maxp);
  }
}

#define heightAt(X, Y) data[(X)+width*(Y)]
const bool HeightField::inside (Vector& point, const Vector& old) const {
  if ((point.x() < origin.x()) || (point.x() >= origin.x() + size.x()) ||
      (point.y() < origin.y()) || (point.y() >= origin.y() + size.y()))
    return false;
  int i = static_cast<int> (((point.x() - origin.x())*width)/size.x());
  int j = static_cast<int> (((point.y() - origin.y())*length)/size.y());
  if (point.z() >= heightAt(i, j))
    return false;

  int i2 = static_cast<int> (((old.x() - origin.x())*width)/size.x());
  int j2 = static_cast<int> (((old.y() - origin.y())*length)/size.y());

  if (((i2 == i) && (j2 == j)) || (heightAt (i, j) < heightAt (i2, j2))) {
                                // both points are in the same cell,
                                // or the new cell is lower than the old one
                                // stop on the surface
    if (old.z() < heightAt (i2, j2)) {
      point = old;
    } else {
      double lambda = (heightAt (i, j)-old.z())/(point.z()-old.z()+EPSILON);
      point = old + lambda*(point-old);
    }
    return true;
  }
                                // the points are in different cells
                                // and the new cell is higher
                                // stop at the cell border
/*  double lambda1 = 1;
  double lambda2 = 1;
#define toX(X) (((X)*size.x())/(double)width+origin.x())
#define toY(Y) (((Y)*size.y())/(double)length+origin.y())
  if (i > i2)
    lambda1 = (toX (i2+1)-old.x())/(point.x()-old.x());
  else if (i < i2) {
    lambda1 = (toX (i2)-old.x())/(point.x()-old.x());*/
    /*cerr << endl << "size: " << size << endl
         << "width: " << width << endl
         << "i2: " << i2 << endl
         << "origin: " << origin << endl
         << "toX(i2): " << toX(i2) << endl
         << "lambda1: " << lambda1 << endl;*/
/*  }
  if (j > j2)
    lambda2 = (toY (j2+1)-old.y())/(point.y()-old.y());
  else if (j < j2)
    lambda2 = (toY (j2)-old.y())/(point.y()-old.y());
  if (lambda1 > lambda2)
    point = old + lambda2*(point-old);
  else
    point = old + lambda1*(point-old);*/
  point = old;
  return true;
}

const void HeightField::adjust (Vector& force, const Vector& point) const {
  if ((point.x() < origin.x()) || (point.x() >= origin.x() + size.x()) ||
      (point.y() < origin.y()) || (point.y() >= origin.y() + size.y()))
    return;
  int i = static_cast<int> (((point.x() - origin.x())*width)/size.x());
  int j = static_cast<int> (((point.y() - origin.y())*length)/size.y());
  if (point.z() < heightAt(i, j) + EPSILON)
    computeFriction (force, Z);
}

const bool HeightField::isInside (const Vector& point) const {
  if ((point.x() < origin.x()) || (point.x() >= origin.x() + size.x()) ||
      (point.y() < origin.y()) || (point.y() >= origin.y() + size.y()))
    return false;
  int i = static_cast<int> (((point.x() - origin.x())*width)/size.x());
  int j = static_cast<int> (((point.y() - origin.y())*length)/size.y());
  return (point.z() < heightAt(i, j));
}

void HeightField::translate (const Vector& v) {
  origin += v;
}

void HeightField::scale (const Vector& v) {
  origin = Vector (origin.x()*v.x(), origin.y()*v.y(), origin.z()*v.z());
  size.x (size.x()*v.x());
  size.y (size.y()*v.y());
  for (int i = length*width-1; i >= 0; i--) data[i]*=v.z();
}
