%{
#include "box.hh"
#include "common.hh"
#include "cylinder.hh"
#include "ground.hh"
#include "hf.hh"
#include "object.hh"
#include "sphere.hh"
#include "vector.hh"
#include "vertex.hh"

#include <fstream>
#include <string>

using namespace std;

int yylex();
void yyerror(char* msg);

class DefaultClothPosition {
public:
  const Vector orig, dir1, dir2;
  DefaultClothPosition (const Vector& orig_, const Vector& d1, const Vector& d2):
    orig(orig_), dir1(d1), dir2(d2) { }
};

class ClothParams {
public:
  int xSize, ySize;
  double dist, d2, strength, s2, curve, weight, hNoise, vNoise, elongation, damping;
  ifstream* read_file;
  DefaultClothPosition* clothPos;
  bool inters;

  ClothParams(): xSize(100), ySize(100), dist(1),
    d2(1.4142135623730951454746218587388284504413604736328125),
    strength(2), s2(1), curve(0.05), weight(0.25),
    hNoise(0.05), vNoise(0.1), elongation(0.1), damping(0.1), read_file(NULL),
    clothPos(NULL), inters(false) {}
  ~ClothParams() { if (clothPos != NULL) delete clothPos; clothPos = NULL; }
} clothParams;

%}

%union {
  Vector* vec;
  double dbl;
  char* str;
  bool boo;
}

//%expect 12

%token <dbl> NUMBER
%token <str> STRING
%token COMMA LCURL RCURL LBRACKET RBRACKET LPAREN RPAREN
%token GLOBAL_SETTINGS STEP OUTPUT MESH MESH2 CYLINDERS WIND VISCOSITY GRAVITY
%token CLOTH XSIZE YSIZE DIST STRENGTH SHEAR STIFF WEIGHT HNOISE VNOISE ELONGATION DAMPING READ_FILE POSITION INTERSECTIONS
%token TRANSLATE ROTATE SCALE ATTACH
%token GROUND BOX CYLINDER SPHERE HF
%token TGA

%type <vec> vector
%type <dbl> num
%type <boo> boolean

%token ON OFF

%token X_VECTOR Y_VECTOR Z_VECTOR 
%left PLUS MINUS
%left TIMES OVER
%left NEG

%%

input:    global_settings cloth objects
        | cloth objects
;

global_settings: GLOBAL_SETTINGS LCURL settings RCURL
;

settings: /* empty */
        | settings setting
;

setting:  STEP num       { step = $2;                }
        | OUTPUT output_type
        | WIND vector    { wind = *$2; delete $2;    }
        | VISCOSITY num  { visc = $2;                }
        | GRAVITY vector { gravity = *$2; delete $2; }
;

output_type:      MESH          { output_type = mesh;           }
                | MESH2         { output_type = mesh2;          }
                | CYLINDERS     { output_type = cylinders;      }
;

cloth: CLOTH LCURL cloth_params 
  {
    if (clothParams.read_file == NULL) {
      if (clothParams.clothPos == NULL)
        cloth = new Cloth (clothParams.xSize, clothParams.ySize, clothParams.dist,
                           clothParams.d2*clothParams.dist,
                           clothParams.strength/clothParams.dist,
                           clothParams.s2/clothParams.dist,
                           clothParams.curve/clothParams.dist,
                           clothParams.weight*clothParams.dist*clothParams.dist,
                           clothParams.hNoise, clothParams.vNoise,
                           clothParams.elongation, clothParams.damping/step,
                           clothParams.inters);
      else
        cloth = new Cloth (clothParams.xSize, clothParams.ySize, clothParams.dist,
                           clothParams.d2*clothParams.dist,
                           clothParams.strength/clothParams.dist,
                           clothParams.s2/clothParams.dist,
                           clothParams.curve/clothParams.dist,
                           clothParams.weight*clothParams.dist*clothParams.dist,
                           clothParams.hNoise, clothParams.vNoise, clothParams.elongation,
                           clothParams.damping/step, clothParams.inters,
                           clothParams.clothPos->orig,
                           clothParams.clothPos->dir1, clothParams.clothPos->dir2);
    } else {
      string s;
      char c;
      while ((clothParams.read_file->good()) && ((c = clothParams.read_file->get()) != '\n'))
        s += c;
      if (!clothParams.read_file->good())
        yyerror ("Unexpected end of file in cloth data\n");
      (*clothParams.read_file) >> clothParams.xSize;
      (*clothParams.read_file) >> clothParams.ySize;
      cloth = new Cloth (*clothParams.read_file, s, clothParams.xSize, clothParams.ySize,
                         clothParams.dist, clothParams.d2*clothParams.dist,
                         clothParams.strength/clothParams.dist,
                         clothParams.s2/clothParams.dist,
                         clothParams.curve/clothParams.dist,
                         clothParams.weight*clothParams.dist*clothParams.dist,
                         clothParams.hNoise, clothParams.vNoise, clothParams.elongation,
                         clothParams.damping/step, clothParams.inters);
      delete clothParams.read_file;
    }
  }
  transforms RCURL
;

cloth_params: /* empty */
                | cloth_params cloth_param
;

cloth_param:      XSIZE num          { clothParams.xSize      = static_cast<int> ($2); }
                | YSIZE num          { clothParams.ySize      = static_cast<int> ($2); }
                | DIST num           { clothParams.dist       = $2;  }
                | STRENGTH num       { clothParams.strength   = $2;  }
                | SHEAR num          { clothParams.s2         = $2;  }
                | STIFF num          { clothParams.curve      = $2;  }
                | WEIGHT num         { clothParams.weight     = $2;  }
                | HNOISE num         { clothParams.hNoise     = $2;  }
                | VNOISE num         { clothParams.vNoise     = $2;  }
                | ELONGATION num     { clothParams.elongation = $2;  }
                | DAMPING num        { clothParams.damping    = $2;  }
                | READ_FILE STRING   { clothParams.read_file = new ifstream ($2); delete $2; }
                | POSITION LCURL vector COMMA vector COMMA vector RCURL
                        {
                          clothParams.clothPos = new DefaultClothPosition (*$3, *$5, *$7);
                          delete $3; delete $5; delete $7;
                        }
                | INTERSECTIONS boolean { clothParams.inters = $2; }
;

transforms: /* empty*/
                | transforms transform
;

transform:        TRANSLATE vector      { cloth->translate (*$2); delete $2; }
                | ROTATE vector         { cloth->rotate (*$2); delete $2;    }
                | ATTACH LCURL num COMMA num COMMA vector RCURL
                        { cloth->attach ($3, $5, *$7); delete $7; }
                | ATTACH LCURL num COMMA num COMMA vector COMMA vector RCURL
                        { cloth->attach ($3, $5, *$7, *$9); delete $7; delete $9 }
;

objects: /* empty */
                | objects object
;

object:           GROUND LCURL num RCURL
                        { world.push_front (new GroundPlane ($3));      }
                | GROUND LCURL num COMMA num RCURL
                        { world.push_front (new GroundPlane ($3, $5));  }
                | BOX LCURL vector COMMA vector RCURL
                        { world.push_front (new Box (*$3, *$5)); delete $3; delete $5; }
                | BOX LCURL vector COMMA vector COMMA num RCURL
                        { world.push_front (new Box (*$3, *$5, $7)); delete $3; delete $5; }
                | CYLINDER LCURL vector COMMA vector COMMA num RCURL
                        { world.push_front (new Cylinder (*$3, *$5, $7)); delete $3; delete $5; }
                | CYLINDER LCURL vector COMMA vector COMMA num COMMA num RCURL
                        { world.push_front (new Cylinder (*$3, *$5, $7, $9)); delete $3; delete $5; }
                | SPHERE LCURL vector COMMA num RCURL
                        { world.push_front (new Sphere (*$3, $5)); delete $3;     }
                | SPHERE LCURL vector COMMA num COMMA num RCURL
                        { world.push_front (new Sphere (*$3, $5, $7)); delete $3; }
                | HF LCURL TGA STRING { world.push_front (new HeightField ($4)); } hf_transforms RCURL
;

hf_transforms: /* empty*/
                | hf_transforms hf_transform
;

hf_transform:     TRANSLATE vector
                        { static_cast<HeightField*> (world.front())->translate (*$2); delete $2; }
                | SCALE vector
                        { static_cast<HeightField*> (world.front())->scale (*$2); delete $2; }
;

vector:   LBRACKET num COMMA num COMMA num RBRACKET
                { $$ = new Vector ($2, $4, $6); }
        | X_VECTOR               { $$ = new Vector (X);                                  }
        | Y_VECTOR               { $$ = new Vector (Y);                                  }
        | Z_VECTOR               { $$ = new Vector (Z);                                  }
        | vector PLUS vector     { $$ = new Vector ((*$1) + (*$3)); delete $1; delete $3;}
        | vector MINUS vector    { $$ = new Vector ((*$1) - (*$3)); delete $1; delete $3;}
        | num TIMES vector       { $$ = new Vector ($1 * (*$3)); delete $3;              }
        | vector TIMES num       { $$ = new Vector ($3 * (*$1)); delete $1;              }
        | vector OVER num        { $$ = new Vector ((1/$3) * (*$1)); delete $1;          }
        | LPAREN vector RPAREN   { $$ = $2;                                              }
        | MINUS vector %prec NEG { $$ = new Vector (-(*$2)); delete $2;                  }
;

num:      NUMBER                { $$ = $1;      }
        | PLUS num %prec NEG    { $$ = $2;      }
        | MINUS num %prec NEG   { $$ = -$2;     }
        | num PLUS num          { $$ = $1 + $3; }
        | num MINUS num         { $$ = $1 - $3; }
        | num TIMES num         { $$ = $1 * $3; }
        | num OVER num          { $$ = $1 / $3; }
        | LPAREN num RPAREN     { $$ = $2;      }
;

boolean:  ON  { $$ = true;      }
        | OFF { $$ = false;     }
        | num { $$ = ($1 == 0); }
;

%%
