#include "common.hh"
#include "object.hh"
#include "utils.hh"
#include "vertex.hh"

#include <list>

using namespace std;

#define weight     owner->weight
#define dist       owner->dist
#define d2         owner->d2
#define strength   owner->strength
#define s2         owner->s2
#define curve      owner->curve
#define elongation owner->elongation
#define maxDist    owner->maxDist
#define maxD2      owner->maxD2
#define damping    owner->damping

Vertex::Vertex(): location_ (), speed(), force(), normal(), newloc(), adjusted(),
                  nadj(), neighbors(), corners(), curvings(), owner((Cloth*)-1),
                  fixed(), attached() {}

Vertex::Vertex (const Vector& iloc, Cloth* iowner):
  location_ (iloc), speed(), force(), normal(), newloc(), adjusted(), nadj(),
  neighbors(), corners(), curvings(), owner (iowner), fixed (false),
  attached(false) {}

void Vertex::addNeighbor (Vertex* n) {
  neighbors.push_front (n);
}

void Vertex::addCorner (Vertex* c) {
  corners.push_front (c);
}

void Vertex::addCurving (Vertex* first, Vertex* second) {
  curvings.push_front (pair<Vertex*, Vertex*> (first, second));
}

void Vertex::computeForce() {
  //if (force.length() > dist) cerr << endl << "Starting: " << force.length();

  // external forces
  if (!attached) {
    Vector norm = (dist/normal.length())*normal;
    force += weight*gravity - damping*speed - (visc*norm.dot (speed - wind))*norm;
  }

  // neighbors
  for (list<Vertex*>::iterator n = neighbors.begin(); n != neighbors.end(); n++) {
    Vector v = location_ - (*n)->location_;
    double l = v.length();
    //Vector f = ((l - dist)*strength*(((l-dist)/dist>elongation)?100:1)/l)*v;
    double lambda;
    if (l <= maxDist)
      lambda = strength*(l-dist);
    else
      lambda = strength*(dist*elongation + (l-maxDist)*2);
    Vector f = (lambda/l)*v;
    force -= f;
    (*n)->force += f;
  }
  //if (force.length() > dist) cerr << endl << "Neighbors: " << force.length();

  // corners
  for (list<Vertex*>::iterator c = corners.begin(); c != corners.end(); c++) {
    Vector v = location_ - (*c)->location_;
    double l = v.length();
    //Vector f = ((l - d2)*s2*(((l-d2)/d2>elongation)?100:1)/l)*v;
    double lambda;
    if (l <= maxD2)
      lambda = s2*(l-d2);
    else
      lambda = s2*(d2*elongation + (l-maxD2)*2);
    Vector f = (lambda/l)*v;
    force -= f;
    (*c)->force += f;
  }
  //if (force.length() > dist) cerr << endl << "Corners: " << force.length();

  // curving
  if (!curvings.empty()) {
    for (list<pair<Vertex*, Vertex*> >::iterator f = curvings.begin();
         f != curvings.end(); f++) {
      Vector v = (f->first->location_ - f->second->location_);
      v *= (curve*(2*dist-v.length())/v.length());
      f->first->force += v;
      f->second->force -= v;
    }
  }
}

void Vertex::adjust() {
  if (attached) return;
  //if (force.length() > dist) cerr << endl << "Before adjusting: " << force.length();
  for (list<Object*>::iterator o = world.begin(); o != world.end(); o++)
    (*o)->adjust (force, location_);
  //if (force.length() > dist) cerr << endl << "After adjusting: " << force.length();
}

void Vertex::move() {
  if (attached) return;
  adjust();
  speed += step*force;
  newloc = location_ + step*speed;
  //newloc = location_ + step*speed + (0.5*step*step/weight)*force;
  fixed = false;
  for (list<Object*>::iterator o = world.begin(); o != world.end(); o++)
    fixed |= (*o)->inside (newloc, location_);
  //speed += (step/weight)*force;
  adjusted = O;
  nadj = 0;
}

void Vertex::doMove() {
  if (attached) {
    location_ += step*speed;
    return;
  }
  //fixed = false;
  //for (list<Object*>::iterator o = world.begin(); o != world.end(); o++)
    //fixed |= (*o)->inside (newloc, location_);
  if (fixed)
    speed = O;
  location_ = newloc;
}

void Vertex::adjustLength() {
  Vector tmp;
  // Structural springs
  for (list<Vertex*>::iterator i = neighbors.begin(); i != neighbors.end(); i++)
    if (((!attached)||(!(*i)->attached)) &&
        ((tmp = ((*i)->newloc-newloc)).l2() > maxDist*maxDist)) {
      if (attached) {
        (*i)->adjusted += newloc + (maxDist/tmp.length())*tmp;
        (*i)->nadj++;
      } else if ((*i)->attached) {
        adjusted += (*i)->newloc - (maxDist/tmp.length())*tmp;
        nadj++;
      } else {
        Vector l = 0.5*(newloc+(*i)->newloc);
        Vector u = tmp.normalize();
        adjusted += l - (maxDist/2.0)*u;
        (*i)->adjusted += l + (maxDist/2.0)*u;
        (*i)->nadj++;
        nadj++;
      }
    }
  // Shear springs
  for (list<Vertex*>::iterator i = corners.begin(); i != corners.end(); i++)
    if (((!attached)||(!(*i)->attached)) &&
        ((tmp = ((*i)->newloc-newloc)).l2() > maxD2*maxD2)) {
      if (attached) {
        (*i)->adjusted += newloc + (maxD2/tmp.length())*tmp;
        (*i)->nadj++;

      } else if ((*i)->attached) {
        adjusted += (*i)->newloc - (maxD2/tmp.length())*tmp;
        nadj++;
      } else {
        Vector l = 0.5*(newloc+(*i)->newloc);
        Vector u = tmp.normalize();
        adjusted += l - (maxD2/2.0)*u;
        (*i)->adjusted += l + (maxD2/2.0)*u;
        (*i)->nadj++;
        nadj++;
      }
    }
}

void Vertex::attach (const Vector& where, const Vector& whatSpeed) {
  attached = fixed = true;
  newloc = location_ = where;
  speed = whatSpeed;
}

void Vertex::endAdjust() {
  if (attached) return;
  if (nadj != 0) {
    Vector tmp = (1/(double)nadj) * adjusted;
    bool trouble = false;
    for (list<Object*>::iterator o = world.begin(); (!trouble)&&(o != world.end()); o++)
      trouble |= (*o)->isInside (tmp);
    if (!trouble) {
      newloc = tmp;
      speed = (1/step)*(newloc - location_);
    } /*else {
      trouble = false;
      for (list<Object*>::iterator o = world.begin(); (!trouble)&&(o != world.end()); o++)
        trouble |= (*o)->isInside (newloc);
      if (trouble)
        cerr << endl << newloc << " =/=> " << tmp;
    }*/
  }
}

void Vertex::intersect (Vertex& a, Vertex& b, Vertex& c) {
  nIntersectChecks++;
  Vector ap = location_ - a.location_;
  Vector ab = b.location_ - a.location_;
  Vector ac = c.location_ - a.location_;
  Vector va = a.newloc - a.location_;
  Vector vp = newloc - location_;
  Vector vpva = vp - va;
  Vector vb = b.newloc - b.location_;
  Vector vbva = vb - va;
  Vector vc = c.newloc - c.location_;
  Vector vcva = vc - va;
  double poly[4];
  poly[0] = ap.dot (ab.cross (ac));
  Vector tmp1 = vbva.cross (ac) + ab.cross (vcva);
  poly[1] = ap.dot (tmp1) + ab.cross (ac).dot (vpva);
  Vector tmp2 = vbva.cross (vcva);
  poly[2] = vpva.dot (tmp1) + ap.dot (tmp2);
  poly[3] = vpva.dot (tmp2);
  double result[3];
  int n = solveCubic (poly, result); // Find possible times of intersection
  for (int i = 0; i < n; i++) {
    double t = result[i];
    if (t > 1.0) return;        // too late
    if (t < 0.0) continue;      // in the past
                                // Maybe we have an intersection, let's check
    Vector at = a.location_ + t*va;
    Vector abt = b.location_ + t*vb - at;
    Vector act = c.location_ + t*vc - at;
    Vector apt = location_ + t*vp - at;

    double dz = abt.x()*act.y() - abt.y()*act.x();
    double adz = fabs (dz);
    double dy = abt.x()*act.z() - abt.z()*act.x();
    double ady = fabs (dy);
    double dx = abt.y()*act.z() - abt.z()*act.y();
    double adx = fabs (dx);
    double u, v;
    if ((adz < EPSILON) && (ady < EPSILON) && (adx < EPSILON)) continue;
    if ((adz >= ady) && (adz >= adx)) {
      u = (apt.x()*act.y() - apt.y()*act.x())/dz;
      v = (abt.x()*apt.y() - abt.y()*apt.x())/dz;
    } else if (ady >= adx) {
      u = (apt.x()*act.z() - apt.z()*act.x())/dy;
      v = (abt.x()*apt.z() - abt.z()*apt.x())/dy;
    } else {
      u = (apt.y()*act.z() - apt.z()*act.y())/dx;
      v = (abt.y()*apt.z() - abt.z()*apt.y())/dx;
    }

    if ((u >= 0) && (v >= 0) && ((u+v) <= 1)) { // there is an intersection!
/*      cerr << "current: " << t << endl
           << "AB: " << abt << endl
           << "AC: " << act << endl
           << "AP: " << apt << endl
           << "u: " << u << ", v: " << v << endl
           << "dx: " << dx << ", dy: " << dy << ", dz: " << dz << endl
           << "*********************************" << endl;*/
      if (t >= EPSILON) {
        nIntersectFound++;
        t *= 0.9;
//Vector zp = newloc, za = a.newloc, zb = b.newloc, zc = c.newloc;
/*Vector tmp = location_ + t*vp;
Vector tmp2 = tmp - u*abt - v*act;
if (vp.dot (tmp2) < 0) {
  newloc = tmp;
}
Vector tmp3 = va + u*(vb-va) + v*(vc-va);
if (tmp3.dot (tmp2) > 0) {
  a.newloc = a.location_ + t*va;
  b.newloc = b.location_ + t*vb;
  c.newloc = c.location_ + t*vc;
}*/

        newloc = location_ + t*vp;
        a.newloc = a.location_ + t*va;
        b.newloc = b.location_ + t*vb;
        c.newloc = c.location_ + t*vc;

        a.speed = b.speed = c.speed = speed = 0.25*(speed + a.speed + b.speed + c.speed);

        /*if (intersect2 (a, b, c)) {
          cerr << endl << poly[0] << " + " << poly[1] << "*x + " << poly[2] << "*x^2 + " << poly[3] << "*x^3"
               << endl << n << ": " << result[0] << ", " << result[1] << ", " << result[2]
               << endl << "t: " << t
               << endl << "P: " << location_ << " => " << zp << " / " << newloc
               << endl << "A: " << a.location_ << " => " << za << " / " << a.newloc
               << endl << "B: " << b.location_ << " => " << zb << " / " << b.newloc
               << endl << "C: " << c.location_ << " => " << zc << " / " << c.newloc
               << endl << "u: " << u << ", v:" << v
               << endl;
          exit (1);
        }*/
      } else {
        newloc = location_;
        a.newloc = a.location_;
        b.newloc = b.location_;
        c.newloc = c.location_;
      }
      return;
    }
  }
}
